"""
@Author = 'Michael Stanley'

============ Change Log ============
2025-Jan-13 = Created. 

============ License ============
Copyright (C) 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import pytest
from wmul_rivendell.LoadCartDataDump import RivendellCart, CartType
from wmul_rivendell.DatabaseStatistics import RivendellGroupStatistics, StatisticsLimits, _remove_outliers
from wmul_test_utils import make_namedtuple, generate_true_false_matrix_from_list_of_strings


@pytest.fixture(scope="function")
def setup_rivendellgroupstatistics():
    carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='perferendis enim harum repudiandae', artist='nisi ipsum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:22', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='sapiente praesentium voluptates expedita', artist='tenetur assumenda', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:28', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='perferendis laudantium dolorum vel', artist='veritatis debitis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:43', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='minus esse tempore occaecati', artist='assumenda odio', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:35', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='placeat at pariatur mollitia', artist='possimus ut', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:59', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='omnis rem dolore eius', artist='accusantium officia', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:13', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='doloremque nostrum fugiat nisi', artist='voluptatum quibusdam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:47', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='dolor dignissimos minima ipsa', artist='vero rem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='5:53', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='odio dolor sint ex', artist='minus aut', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='1:59', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='ipsum repellendus aspernatur qui', artist='voluptatibus qui', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='provident quod repellendus architecto', artist='ad amet', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:45', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='nobis totam architecto molestias', artist='soluta quasi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:49', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='ducimus cum dolore accusantium', artist='dolores iusto', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:13', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='iusto dolores ab dolore', artist='beatae excepturi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:27', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='est eum molestiae eveniet', artist='veniam modi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='5:08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='dolores ducimus odit doloremque', artist='ab possimus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:40', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='enim veniam a aspernatur', artist='nulla at', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:39', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='eligendi commodi ex corrupti', artist='rem nobis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:31', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='ex beatae facilis alias', artist='error eligendi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:37', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='sunt veniam ab dicta', artist='minus sed', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:02', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='cumque excepturi eum nostrum', artist='amet aperiam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='10:53', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='maiores tenetur dolor sequi', artist='sequi quae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:17', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='aperiam consequatur magnam numquam', artist='ut pariatur', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='0:14', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='similique doloribus molestias adipisci', artist='commodi vitae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:32', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='dolorum pariatur esse quia', artist='recusandae mollitia', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:04', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
    ]

    stats_limits = StatisticsLimits(smallest_stdev=15)

    group_name = "EXPLICABO"
    group_stats = RivendellGroupStatistics(group_name=group_name, songs_in_group=carts, stats_limits=stats_limits)

    expected_number_of_songs = 25
    expected_shortest_song_length = 14 # 0:15
    expected_longest_song_length = 653 # 10:53

    expected_lower_outlier_limit = 44 # 44.5 rounded by python 
    expected_upper_outlier_limit = 416 # 416.5 rounded by python
    expected_outlier_limits = (expected_lower_outlier_limit, expected_upper_outlier_limit)

    expected_mean = 229 # 229.217 rounded
    expected_stdev = 57 # 56.8866 rounded

    expected_lower_bound = 150 # 143.5 rounded to the nearest 15 seconds
    expected_upper_bound = 405 # 400.0 rounded to the nearest 15 seconds

    expected_number_of_songs_shorter_than_lower_bound = 3
    expected_number_of_songs_longer_than_upper_bound = 1

    expected_percentage_of_songs_excluded = 16.0

    return make_namedtuple(
        "setup_rivendellgroupstatistics",
        group_name=group_name,
        group_stats=group_stats,
        expected_number_of_songs=expected_number_of_songs,
        expected_shortest_song_length=expected_shortest_song_length,
        expected_longest_song_length=expected_longest_song_length,
        expected_outlier_limits=expected_outlier_limits,
        expected_mean=expected_mean,
        expected_stdev=expected_stdev,
        expected_lower_bound=expected_lower_bound,
        expected_upper_bound=expected_upper_bound,
        expected_number_of_songs_shorter_than_lower_bound=expected_number_of_songs_shorter_than_lower_bound,
        expected_number_of_songs_longer_than_upper_bound=expected_number_of_songs_longer_than_upper_bound,
        expected_percentage_of_songs_excluded=expected_percentage_of_songs_excluded
    )

def test_group_name_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.group_name == setup_rivendellgroupstatistics.group_name

def test_number_of_songs_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.number_of_songs == setup_rivendellgroupstatistics.expected_number_of_songs
    
def test_shortest_song_length_in_this_group_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.shortest_song_length == setup_rivendellgroupstatistics.expected_shortest_song_length

def test_longest_song_length_in_this_group_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.longest_song_length == setup_rivendellgroupstatistics.expected_longest_song_length
    
def test_outlier_limits_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.outlier_limits == setup_rivendellgroupstatistics.expected_outlier_limits

def test_mean_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.mean == setup_rivendellgroupstatistics.expected_mean

def test_stdev_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.stdev == setup_rivendellgroupstatistics.expected_stdev

def test_lower_bound_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.lower_bound == setup_rivendellgroupstatistics.expected_lower_bound

def test_upper_bound_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert group_stats.upper_bound == setup_rivendellgroupstatistics.expected_upper_bound

def test_number_of_songs_shorter_than_lower_bound_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert (group_stats.number_of_songs_shorter_than_lower_bound == 
        setup_rivendellgroupstatistics.expected_number_of_songs_shorter_than_lower_bound)

def test_number_of_songs_longer_than_upper_bound_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert (group_stats.number_of_songs_longer_than_upper_bound == 
        setup_rivendellgroupstatistics.expected_number_of_songs_longer_than_upper_bound)
    
def test_percentage_excluded_correct(setup_rivendellgroupstatistics):
    group_stats = setup_rivendellgroupstatistics.group_stats
    assert (group_stats.percentage_of_songs_excluded == 
            setup_rivendellgroupstatistics.expected_percentage_of_songs_excluded)

def test_stdev_is_nan():
    """
    The standard deviation of a single item is Not a Number (NaN). RivendellGroupStatistics.stdev should be 0 in 
    this case.
    """    
    carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='perferendis enim harum repudiandae', artist='nisi ipsum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:22', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
    ]

    stats_limits = StatisticsLimits()

    group_name = "EXPLICABO"
    group_stats = RivendellGroupStatistics(group_name=group_name, songs_in_group=carts, stats_limits=stats_limits)

    assert group_stats.stdev == 0


def test_lower_bound_is_less_than_zero():
    """
    If the calculated lower bound is below zero, RivendellGroupStatistics.lower_bound should be 0.
    """
    carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='quo architecto asperiores minus', artist='ullam enim', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='dignissimos illum eius maxime', artist='id repellat', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='fugiat sunt in tenetur', artist='amet quam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='voluptates praesentium eveniet tempora', artist='quod culpa', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':07', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='cum ipsam labore harum', artist='ullam praesentium', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':10', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='alias ratione similique laudantium', artist='dolor iure', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':07', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='tempora hic quos eius', artist='sed ullam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':07', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='reprehenderit minima quidem veritatis', artist='vero laboriosam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='veritatis debitis tempora deserunt', artist='doloremque exercitationem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':06', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='ea exercitationem officiis ipsum', artist='occaecati totam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':07', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='minus corrupti blanditiis facilis', artist='quisquam officia', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':04', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='tempore corrupti saepe alias', artist='delectus minus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='at occaecati itaque sint', artist='sint nisi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='ducimus id asperiores eum', artist='optio ratione', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='consectetur illo tempore alias', artist='perspiciatis maxime', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='consectetur incidunt nihil alias', artist='cupiditate culpa', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='amet ipsam veniam suscipit', artist='adipisci at', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='delectus aliquam assumenda autem', artist='harum delectus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':11', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='VOLUPTATIB', title='odit ex hic molestiae', artist='debitis voluptas', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
    ]

    stats_limits = StatisticsLimits()

    group_name = "VOLUPTATIB"  
    group_stats = RivendellGroupStatistics(group_name=group_name, songs_in_group=carts, stats_limits=stats_limits)

    assert group_stats.lower_bound == 0


def test_remove_outliers():
    import numpy as np

    carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='atque numquam neque animi', artist='animi voluptas', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:22', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='quos doloremque nulla eligendi', artist='totam voluptatem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:28', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='unde quidem quam debitis', artist='adipisci qui', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:43', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='explicabo commodi maxime nobis', artist='hic veniam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:35', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='architecto consequuntur hic occaecati', artist='aut odit', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:59', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='enim natus expedita in', artist='amet similique', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:13', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='vero labore soluta fugiat', artist='nesciunt incidunt', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:47', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='quae repellat placeat quas', artist='quam illum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='5:53', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='culpa sunt velit itaque', artist='sed quaerat', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='1:59', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='quia possimus assumenda libero', artist='eos itaque', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='voluptas minima exercitationem officiis', artist='quam et', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:45', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='quod dolorem ut illum', artist='culpa impedit', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:49', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='tempore doloremque consectetur occaecati', artist='totam aliquam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:13', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='doloremque quae beatae eum', artist='sint odio', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:27', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='deserunt aspernatur quod perferendis', artist='provident nobis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='5:08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='modi earum sequi quod', artist='excepturi velit', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:40', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='sit dignissimos quisquam reiciendis', artist='necessitatibus incidunt', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:39', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='sed excepturi porro aspernatur', artist='nisi minus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:31', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='pariatur blanditiis sunt corrupti', artist='unde dicta', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:37', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='eius illum consequuntur aperiam', artist='voluptatem cum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:02', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='natus sunt laborum esse', artist='repellendus aut', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='10:53', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='distinctio provident similique placeat', artist='sed veniam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:17', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='ullam magni commodi quaerat', artist='placeat rerum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='0:14', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='quaerat minus unde aliquam', artist='voluptatibus ex', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:32', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='porro ad ipsum repudiandae', artist='explicabo eum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:04', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    ]

    times_of_this_group = np.array([this_item.length_in_seconds() for this_item in carts])

    expected_lower_outlier_limit = 44.5 
    expected_upper_outlier_limit = 416.5 

    stats_limits = StatisticsLimits()

    outliers_excluded, lower_limit, upper_limit = _remove_outliers(
        times_of_this_group=times_of_this_group, 
        stats_limits=stats_limits
    )
    
    carts_without_outliers = [
        RivendellCart(cart_number='100000', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title="There's a Star", artist='Ash', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0001', conductor='', song_id='', user_defined='', description="There's a Star", outcue='', filename='100000_001.wav', length='4:22', start_point='0', end_point='262000', segue_start_point='260000', segue_end_point='262000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='10000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100001', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Six Feet Under', artist='No Doubt', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0002', conductor='', song_id='', user_defined='', description='Six Feet Under', outcue='', filename='100001_001.wav', length='2:28', start_point='0', end_point='148000', segue_start_point='146000', segue_end_point='148000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='6000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100002', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Portions for Foxes', artist='Rilo Kiley', album='', year='2004', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0003', conductor='', song_id='', user_defined='', description='Portions for Foxes', outcue='', filename='100002_001.wav', length='4:43', start_point='0', end_point='283000', segue_start_point='280000', segue_end_point='283000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='34000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s'),
        RivendellCart(cart_number='100003', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Maybe Tomorrow', artist='Stereophonics', album='', year='2004', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0004', conductor='', song_id='', user_defined='', description='Maybe Tomorrow', outcue='', filename='100003_001.wav', length='4:35', start_point='0', end_point='275000', segue_start_point='271000', segue_end_point='275000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='30000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100004', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Existentialism on Prom Night', artist='Straylight Run', album='', year='2004', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0005', conductor='', song_id='', user_defined='', description='Existentialism on Prom Night', outcue='', filename='100004_001.wav', length='3:59', start_point='0', end_point='239000', segue_start_point='234000', segue_end_point='239000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='28000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),

        RivendellCart(cart_number='100005', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Your Ex-Lover is Dead', artist='Stars', album='', year='2004', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0006', conductor='', song_id='', user_defined='', description='Your Ex-Lover is Dead', outcue='', filename='100005_001.wav', length='4:13', start_point='0', end_point='253000', segue_start_point='251000', segue_end_point='253000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='0', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100006', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Neighborhood #1 (Tunnels)', artist='Arcade Fire', album='', year='2004', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0007', conductor='', song_id='', user_defined='', description='Neighborhood #1 (Tunnels)', outcue='', filename='100006_001.wav', length='4:47', start_point='0', end_point='287000', segue_start_point='285000', segue_end_point='287000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='25000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100007', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='How to Disappear Completely', artist='Radiohead', album='', year='2000', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0008', conductor='', song_id='', user_defined='', description='How to Disappear Completely', outcue='', filename='100007_001.wav', length='5:53', start_point='0', end_point='353000', segue_start_point='351000', segue_end_point='353000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='41000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100008', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Pin', artist='Yeah Yeah Yeahs', album='', year='2003', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0009', conductor='', song_id='', user_defined='', description='Pin', outcue='', filename='100008_001.wav', length='1:59', start_point='0', end_point='119000', segue_start_point='117000', segue_end_point='119000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='22000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100009', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Juxtapozed with U', artist='Super Furry Animals', album='', year='2001', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0010', conductor='', song_id='', user_defined='', description='Juxtapozed with U', outcue='', filename='100009_001.wav', length='3:09', start_point='0', end_point='189000', segue_start_point='188000', segue_end_point='189000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='25000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        
        RivendellCart(cart_number='100010', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='New Slang', artist='The Shins', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0011', conductor='', song_id='', user_defined='', description='New Slang', outcue='', filename='100010_001.wav', length='3:45', start_point='0', end_point='225000', segue_start_point='221000', segue_end_point='225000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='36000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100011', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='War on War', artist='Wilco', album='', year='2002', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0012', conductor='', song_id='', user_defined='', description='War on War', outcue='', filename='100011_001.wav', length='3:49', start_point='0', end_point='229000', segue_start_point='227000', segue_end_point='229000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='50000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100012', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Drift Away', artist='Uncle Kracker', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0013', conductor='', song_id='', user_defined='', description='Drift Away', outcue='', filename='100012_001.wav', length='4:13', start_point='0', end_point='253000', segue_start_point='246000', segue_end_point='253000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='12000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100013', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Neighborhood #2 (Laika)', artist='Arcade Fire', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0014', conductor='', song_id='', user_defined='', description='Neighborhood #2 (Laika)', outcue='', filename='100013_001.wav', length='3:27', start_point='0', end_point='207000', segue_start_point='205000', segue_end_point='207000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='23000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100014', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Neighborhood #3 (Power Out)', artist='Arcade Fire', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0015', conductor='', song_id='', user_defined='', description='Neighborhood #3 (Power Out)', outcue='', filename='100014_001.wav', length='5:08', start_point='0', end_point='308000', segue_start_point='308000', segue_end_point='308000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='15000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        
        RivendellCart(cart_number='100015', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Neighborhood #4 (7 Kettles)', artist='Arcade Fire', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0016', conductor='', song_id='', user_defined='', description='Neighborhood #4 (7 Kettles)', outcue='', filename='100015_001.wav', length='4:40', start_point='0', end_point='280000', segue_start_point='272000', segue_end_point='280000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='22000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100016', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='5/4', artist='Gorillaz', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0017', conductor='', song_id='', user_defined='', description='5/4', outcue='', filename='100016_001.wav', length='2:39', start_point='0', end_point='159000', segue_start_point='156000', segue_end_point='159000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='12000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100017', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Cute Without the E', artist='Taking Back Sunday', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0018', conductor='', song_id='', user_defined='', description='Cute Without the E', outcue='', filename='100017_001.wav', length='3:31', start_point='0', end_point='211000', segue_start_point='207000', segue_end_point='211000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='5000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100018', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Pinch Me', artist='Barenaked Ladies', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0019', conductor='', song_id='', user_defined='', description='Pinch Me', outcue='', filename='100018_001.wav', length='4:37', start_point='0', end_point='277000', segue_start_point='270000', segue_end_point='277000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='13000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100019', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Move Your Feet', artist='Junior Senior', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0020', conductor='', song_id='', user_defined='', description='Move Your Feet', outcue='', filename='100019_001.wav', length='3:02', start_point='0', end_point='182000', segue_start_point='180000', segue_end_point='182000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='6000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),

        RivendellCart(cart_number='100021', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Everything Hits At Once', artist='Spoon', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0022', conductor='', song_id='', user_defined='', description='Everything Hits At Once', outcue='', filename='100021_001.wav', length='3:17', start_point='0', end_point='197000', segue_start_point='195000', segue_end_point='197000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='17000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100023', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Call of the Yeti', artist='The Mighty Boosh', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0024', conductor='', song_id='', user_defined='', description='Call of the Yeti', outcue='', filename='100023_001.wav', length='2:32', start_point='0', end_point='152000', segue_start_point='148000', segue_end_point='152000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='0', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s|Flashback'),
        RivendellCart(cart_number='100024', cut_number='1', type=CartType.Audio, group_name='ALTERNATIV', title='Spiky Seeds', artist='Pillows', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='Imported from WOAFR: A00/0025', conductor='', song_id='', user_defined='', description='Spiky Seeds', outcue='', filename='100024_001.wav', length='3:04', start_point='0', end_point='184000', segue_start_point='180000', segue_end_point='184000', hook_start_point='-1', hook_end_point='-1', talk_start_point='0', talk_end_point='20000', fadeup_point='-1', fadedown_point='-1', sched_codes='2000s')
    ]

    times_without_outliers = np.array([this_item.length_in_seconds() for this_item in carts_without_outliers])
    times_without_outliers.sort()

    assert (outliers_excluded == times_without_outliers).all()
    assert lower_limit == expected_lower_outlier_limit
    assert upper_limit == expected_upper_outlier_limit


def test_remove_outliers_not_enough_population():
    """
    If the population is < 4, then the remove_outliers function should return the list unchanged.
    The lower limit should be 0 and the upper limit should be 86,399 (23 hours, 59 minutes, 59 seconds).
    """
    import numpy as np

    carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='atque numquam neque animi', artist='animi voluptas', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:22', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='quos doloremque nulla eligendi', artist='totam voluptatem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:28', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='unde quidem quam debitis', artist='adipisci qui', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:43', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
    ]

    times_of_this_group = np.array([this_item.length_in_seconds() for this_item in carts])

    expected_lower_outlier_limit = 0 
    expected_upper_outlier_limit = 86_399 # 23:59:59

    stats_limits = StatisticsLimits()

    outliers_excluded, lower_limit, upper_limit = _remove_outliers(
        times_of_this_group=times_of_this_group, 
        stats_limits=stats_limits
    )
    
    assert (outliers_excluded == times_of_this_group).all()
    assert lower_limit == expected_lower_outlier_limit
    assert upper_limit == expected_upper_outlier_limit


def test_remove_ouliers_too_small_stdev():
    """
    If the standard deviation of the population is < 15 seconds, don't try to calculate outliers.
    """
    import numpy as np

    carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='magni voluptatum saepe architecto', artist='recusandae temporibus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='ipsum doloremque eveniet odio', artist='repellat rem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='doloremque veniam ex a', artist='praesentium delectus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':04', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='quas accusantium reiciendis illo', artist='quibusdam delectus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='a fugit natus maiores', artist='nemo dolor', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':12', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='eum nihil eaque autem', artist='reprehenderit sapiente', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='a suscipit labore eos', artist='hic ut', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='nostrum cupiditate eaque maiores', artist='consequuntur voluptatum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='voluptate quis animi rerum', artist='repudiandae libero', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':17', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='quos perspiciatis quasi corporis', artist='ducimus fugiat', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='doloremque sit eligendi vitae', artist='ut omnis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='ex quisquam laborum natus', artist='autem sequi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':06', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='mollitia eos id architecto', artist='adipisci voluptas', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':10', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='beatae ipsa impedit illo', artist='repellat impedit', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':10', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='illum hic nostrum quia', artist='culpa cum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':15', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='nemo accusamus aliquam pariatur', artist='architecto facilis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='dignissimos accusamus quod corrupti', artist='molestiae ipsa', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':11', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='molestiae fugit veritatis suscipit', artist='numquam ullam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='consequatur officia nostrum cupiditate', artist='rem distinctio', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='AMC 9/1/2013', description='', outcue='', filename='', length=':13', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='alias magnam deleniti sunt', artist='maxime facere', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='AMC 8/20/2013', description='', outcue='', filename='', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
    ]

    times_of_this_group = np.array([this_item.length_in_seconds() for this_item in carts])

    expected_lower_outlier_limit = 0 
    expected_upper_outlier_limit = 86_399 # 23:59:59

    stats_limits = StatisticsLimits()

    outliers_excluded, lower_limit, upper_limit = _remove_outliers(
        times_of_this_group=times_of_this_group, 
        stats_limits=stats_limits
    )
    
    assert (outliers_excluded == times_of_this_group).all()
    assert lower_limit == expected_lower_outlier_limit
    assert upper_limit == expected_upper_outlier_limit


def test_nearest_15():
    test_cases = [
        (0, 0),
        (1, 0),
        (7, 0),
        (8, 15),
        (22, 15),
        (23, 30),
        (30, 30)
    ]

    for test_value, expected_result in test_cases:
        result = RivendellGroupStatistics._nearest_15(test_value)
        assert result == expected_result


def test_main_calculations_small_stdev():
    carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='magni voluptatum saepe architecto', artist='recusandae temporibus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='ipsum doloremque eveniet odio', artist='repellat rem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='doloremque veniam ex a', artist='praesentium delectus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':04', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='quas accusantium reiciendis illo', artist='quibusdam delectus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='a fugit natus maiores', artist='nemo dolor', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':12', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='eum nihil eaque autem', artist='reprehenderit sapiente', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='a suscipit labore eos', artist='hic ut', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='nostrum cupiditate eaque maiores', artist='consequuntur voluptatum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='voluptate quis animi rerum', artist='repudiandae libero', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':17', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='quos perspiciatis quasi corporis', artist='ducimus fugiat', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='doloremque sit eligendi vitae', artist='ut omnis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':05', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='ex quisquam laborum natus', artist='autem sequi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':06', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='mollitia eos id architecto', artist='adipisci voluptas', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':10', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='beatae ipsa impedit illo', artist='repellat impedit', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':10', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='illum hic nostrum quia', artist='culpa cum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':15', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='nemo accusamus aliquam pariatur', artist='architecto facilis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='dignissimos accusamus quod corrupti', artist='molestiae ipsa', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':11', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='molestiae fugit veritatis suscipit', artist='numquam ullam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length=':09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='consequatur officia nostrum cupiditate', artist='rem distinctio', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='AMC 9/1/2013', description='', outcue='', filename='', length=':13', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='DOLORES', title='alias magnam deleniti sunt', artist='maxime facere', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='AMC 8/20/2013', description='', outcue='', filename='', length=':08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
     ]

    stats_limits = StatisticsLimits()

    group_stats = RivendellGroupStatistics("DOLORES", carts, stats_limits)

    assert group_stats.lower_bound == 0
    assert group_stats.number_of_songs_shorter_than_lower_bound == 0
    assert group_stats.upper_bound == 86_399
    assert group_stats.number_of_songs_longer_than_upper_bound == 0
    assert group_stats.percentage_of_songs_excluded == 0


to_pandas_params, to_pandas_ids = \
    generate_true_false_matrix_from_list_of_strings(
        "to_pandas",
        [
            "write_full_statistics"
        ]

    )

@pytest.mark.parametrize("write_full_statistics", to_pandas_params, ids=to_pandas_ids)
def test_to_pandas_correct(write_full_statistics):
    carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='perferendis enim harum repudiandae', artist='nisi ipsum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:22', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='sapiente praesentium voluptates expedita', artist='tenetur assumenda', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:28', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='perferendis laudantium dolorum vel', artist='veritatis debitis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:43', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='minus esse tempore occaecati', artist='assumenda odio', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:35', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='placeat at pariatur mollitia', artist='possimus ut', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:59', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='omnis rem dolore eius', artist='accusantium officia', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:13', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='doloremque nostrum fugiat nisi', artist='voluptatum quibusdam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:47', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='dolor dignissimos minima ipsa', artist='vero rem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='5:53', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='odio dolor sint ex', artist='minus aut', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='1:59', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='ipsum repellendus aspernatur qui', artist='voluptatibus qui', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:09', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='provident quod repellendus architecto', artist='ad amet', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:45', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='nobis totam architecto molestias', artist='soluta quasi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:49', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='ducimus cum dolore accusantium', artist='dolores iusto', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:13', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='iusto dolores ab dolore', artist='beatae excepturi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:27', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='est eum molestiae eveniet', artist='veniam modi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='5:08', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='dolores ducimus odit doloremque', artist='ab possimus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:40', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='enim veniam a aspernatur', artist='nulla at', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:39', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='eligendi commodi ex corrupti', artist='rem nobis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:31', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='ex beatae facilis alias', artist='error eligendi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='4:37', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='sunt veniam ab dicta', artist='minus sed', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:02', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='cumque excepturi eum nostrum', artist='amet aperiam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='10:53', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='maiores tenetur dolor sequi', artist='sequi quae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:17', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='aperiam consequatur magnam numquam', artist='ut pariatur', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='0:14', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='similique doloribus molestias adipisci', artist='commodi vitae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='2:32', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='dolorum pariatur esse quia', artist='recusandae mollitia', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='3:04', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    ]

    stats_limits = StatisticsLimits()

    group_name = "EXPLICABO"
    group_stats = RivendellGroupStatistics(group_name=group_name, songs_in_group=carts, stats_limits=stats_limits)

    import pandas as pd

    if write_full_statistics:
        expected_pandas_series = pd.Series(
            {
                "Number of Songs": 25,
                "Shortest Song Length": "0:00:14",
                "Longest Song Length": "0:10:53",
                "Outlier Limits": ("0:00:44", "0:06:56"),
                "Mean": "0:03:49",
                "Standard Deviation": "0:00:57",
                "Lower Bound": "0:02:30",
                "Number of Songs < Lower Bound": 3,
                "Upper Bound": "0:06:45",
                "Number of Songs > Upper Bound": 1,
                "Percent of Songs Excluded": 16.0
            }
        )
    else:
        expected_pandas_series = pd.Series(
            {
                "Number of Songs": 25,
                "Number of Songs < Lower Bound": 3,
                "Number of Songs > Upper Bound": 1
            }
        )

    result_series = group_stats.to_pandas_series(write_full_statistics)

    assert (result_series == expected_pandas_series).all()



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2020-Oct-08 = Created.

============ License ============
Copyright (C) 2020, 2023 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import pytest
import wmul_test_utils

from collections import namedtuple
from wmul_rivendell import RivendellAudioImporter


_check_this_file_test_options = namedtuple(
    "_check_this_file_test_options",
    [
        "is_file",
        "previously_seen_and_same_size",
        "for_importer"
    ]
)
_check_this_file_test_params, _check_this_file_test_ids = \
    wmul_test_utils.generate_true_false_matrix_from_namedtuple(_check_this_file_test_options)


@pytest.fixture(scope="function", params=_check_this_file_test_params, ids=_check_this_file_test_ids)
def setup__check_this_file(request, mocker):
    param = request.param

    mock_file_path_is_file = mocker.Mock(return_value=param.is_file)
    mock_file_path = mocker.Mock(
        is_file=mock_file_path_is_file
    )

    mock_datetime_now = "mock_datetime_now"
    mock_source_path = "mock_source_path"

    mock_previously_seen_version = "mock_previously_seen_version"
    mock_previously_seen_files = {
        mock_file_path: mock_previously_seen_version
    }

    mock_this_version = "mock_this_version"
    mock_file_information = mocker.Mock(return_value=mock_this_version)
    mocker.patch("wmul_rivendell.RivendellAudioImporter.FileInformation",
                 mock_file_information)

    def mock__previous_seen_and_same_size_function(psv, tv):
        return param.previously_seen_and_same_size
    mock__previous_seen_and_same_size = mocker.Mock(side_effect=mock__previous_seen_and_same_size_function)
    mocker.patch("wmul_rivendell.RivendellAudioImporter._previous_seen_and_same_size",
                 mock__previous_seen_and_same_size)

    def mock__for_importer_or_seen_this_time_function(psv, tv):
        if param.for_importer:
            return None, tv
        else:
            return psv, None
    mock__for_importer_or_seen_this_time = mocker.Mock(side_effect=mock__for_importer_or_seen_this_time_function)
    mocker.patch("wmul_rivendell.RivendellAudioImporter._for_importer_or_seen_this_time",
                 mock__for_importer_or_seen_this_time)

    result_version, result_for_importer = RivendellAudioImporter._check_this_file(
        mock_previously_seen_files, mock_file_path, mock_datetime_now, mock_source_path
    )

    return wmul_test_utils.make_namedtuple(
        "setup__check_this_file",
        param=param,
        mock_file_path_is_file=mock_file_path_is_file,
        mock_file_path=mock_file_path,
        mock_datetime_now=mock_datetime_now,
        mock_source_path=mock_source_path,
        mock_previously_seen_version=mock_previously_seen_version,
        mock_this_version=mock_this_version,
        mock_file_information=mock_file_information,
        mock__previous_seen_and_same_size=mock__previous_seen_and_same_size,
        mock__for_importer_or_seen_this_time=mock__for_importer_or_seen_this_time,
        result_version=result_version,
        result_for_importer=result_for_importer
    )


def test_file_path_is_file_called_correctly(setup__check_this_file):
    setup__check_this_file.mock_file_path_is_file.assert_called_once_with()


def test_file_information_called_correctly(setup__check_this_file):
    if setup__check_this_file.param.is_file:
        setup__check_this_file.mock_file_information.assert_called_once_with(
            file_path=setup__check_this_file.mock_file_path,
            timestamp=setup__check_this_file.mock_datetime_now,
            source_path=setup__check_this_file.mock_source_path
        )
    else:
        setup__check_this_file.mock_file_information.assert_not_called()


def test__previous_seen_called_correctly(setup__check_this_file):
    if setup__check_this_file.param.is_file:
        setup__check_this_file.mock__previous_seen_and_same_size.assert_called_once_with(
            setup__check_this_file.mock_previously_seen_version,
            setup__check_this_file.mock_this_version
        )
    else:
        setup__check_this_file.mock__previous_seen_and_same_size.assert_not_called()


def test__for_importer_called_correctly(setup__check_this_file):
    if setup__check_this_file.param.is_file:
        if setup__check_this_file.param.previously_seen_and_same_size:
            setup__check_this_file.mock__for_importer_or_seen_this_time.assert_called_once_with(
                setup__check_this_file.mock_previously_seen_version,
                setup__check_this_file.mock_this_version
            )
        else:
            setup__check_this_file.mock__for_importer_or_seen_this_time.assert_not_called()
    else:
        setup__check_this_file.mock__for_importer_or_seen_this_time.assert_not_called()


def test__version_correct(setup__check_this_file):
    if setup__check_this_file.param.is_file:
        if setup__check_this_file.param.previously_seen_and_same_size:
            if setup__check_this_file.param.for_importer:
                assert setup__check_this_file.result_version is None
            else:
                assert setup__check_this_file.result_version == setup__check_this_file.mock_previously_seen_version
        else:
            assert setup__check_this_file.result_version == setup__check_this_file.mock_this_version
    else:
        assert setup__check_this_file.result_version is None


def test_for_importer_correct(setup__check_this_file):
    if setup__check_this_file.param.is_file:
        if setup__check_this_file.param.previously_seen_and_same_size:
            if setup__check_this_file.param.for_importer:
                assert setup__check_this_file.result_for_importer == setup__check_this_file.mock_this_version
            else:
                assert setup__check_this_file.result_for_importer is None
        else:
            assert setup__check_this_file.result_for_importer is None
    else:
        assert setup__check_this_file.result_for_importer is None



"""
@Author = 'Michael Stanley'

I am reasonably certain that this was based on Open Source Radio's 'load-log-skip-to-current-line':
https://github.com/opensourceradio/ram/blob/e112952d87a64d92d564ab9693d37c9e63740607/usr/local/bin/load-log-skip-to-current-time .
However, I did not document my source at the time I originally wrote this.
That module is BSD 2-Clause licensed. 

============ Change Log ============
2023-Jan-20 = Back-insert items into change log.
              Updated description.
              Change license from GPLv2 to GPLv3.

2022-May-09 = Added description and credit.

2021-May-26 = Inline finding the log name from the date.

2020-Jun-26 = Created.

============ License ============
Copyright (C) 2020-2023 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
from dataclasses import dataclass
from datetime import datetime, timedelta
from mysql.connector import Error

import mysql.connector
import subprocess
import wmul_emailer
import wmul_logger

_logger = wmul_logger.get_logger()


@dataclass
class LoadCurrentLogLineArguments:
    log_datetime: datetime
    sql_host: str
    sql_user: str
    sql_pass: str
    sql_database_name: str
    log_name_format: str
    rivendell_host: str
    dry_run: bool
    start_immediately: bool
    days_back: int
    default_log: str
    log_machine: int
    emailer: wmul_emailer.EmailSender


def _get_start_time_in_millis(log_datetime):
    desired_time = log_datetime.time()
    hours_in_millis = desired_time.hour * 3_600_000
    minutes_in_millis = desired_time.minute * 60_000
    seconds_in_millis = desired_time.second * 1000
    return hours_in_millis + minutes_in_millis + seconds_in_millis


def _get_desired_log_line(arguments, desired_start_time_in_millis):
    connection = mysql.connector.connect(
        host=arguments.sql_host,
        database=arguments.sql_database_name,
        user=arguments.sql_user,
        password=arguments.sql_pass
    )
    working_datetime = arguments.log_datetime
    working_days_back = arguments.days_back
    log_name = None
    desire_log_line = None
    try:
        if connection.is_connected():
            cursor = connection.cursor()
            while desire_log_line is None:
                log_name = working_datetime.strftime(arguments.log_name_format)
                desire_log_line = _query_database_for_log_line(
                    log_name=log_name,
                    desired_start_time_in_millis=desired_start_time_in_millis,
                    cursor=cursor
                )
                if desire_log_line is None:
                    if working_days_back > 0:
                        working_datetime -= timedelta(days=1)
                        working_days_back -= 1
                    elif arguments.default_log:
                        log_name = arguments.default_log
                        desire_log_line = _get_default_log(cursor, log_name, desired_start_time_in_millis)
                    else:
                        _log_possibilities_exhausted()
                        break
    except Error as e:
        _logger.error("Error while connecting to MySQL", e)
    finally:
        if connection.is_connected():
            cursor.close()
            connection.close()
            _logger.debug("MySQL connection is closed")
    return desire_log_line, log_name


def _get_default_log(cursor, log_name, desired_start_time_in_millis):
    record = _query_database_for_log_line(
        log_name=log_name,
        desired_start_time_in_millis=desired_start_time_in_millis,
        cursor=cursor
    )
    if record is None:
        _log_possibilities_exhausted()
    return record


def _log_possibilities_exhausted():
    _logger.critical("All log possibilities exhausted.")


def _query_database_for_log_line(log_name, desired_start_time_in_millis, cursor):
    sql_statement = "SELECT count, cart_number, start_time FROM LOG_LINES WHERE log_name = %s AND start_time " \
                    ">= 0 ORDER BY count ASC LIMIT 1;"
    cursor.execute(sql_statement, (log_name,))
    first_line = cursor.fetchone()
    if first_line is None:
        _logger.debug(f"There are no log lines with Log Name: {log_name}")
        return None
    if desired_start_time_in_millis == 0:
        return first_line

    sql_statement = "SELECT count, cart_number, start_time FROM LOG_LINES WHERE log_name = %s AND " \
                    "start_time > 0 AND start_time <= %s ORDER BY count DESC LIMIT 1;"
    cursor.execute(sql_statement, (log_name, desired_start_time_in_millis))
    line_closest_to_time = cursor.fetchone()
    if line_closest_to_time is None:
        return first_line
    else:
        return line_closest_to_time


def _build_rml_command(desired_log_line, log_name, start_immediately, log_machine):
    load_log_command = f"LL {log_machine} {log_name} {desired_log_line[0]} ! "
    if start_immediately:
        start_immediately_command = f"PN {log_machine} !"
    else:
        start_immediately_command = ""
    set_mode_automatic_command = f"PM 2 {log_machine} ! "
    return f"{load_log_command} {start_immediately_command} {set_mode_automatic_command}"


def _send_command_to_rivendell(rml_command, rivendell_host):
    try:
        rml_process = subprocess.run(
            ["rmlsend", f"--to-host={rivendell_host}", rml_command],
            capture_output=True,
            check=True
        )
        _logger.debug(f"Sending command to Rivendell succeeded. {rml_process}")
        return str(rml_process)
    except subprocess.CalledProcessError as e:
        _logger.critical("Sending command to Rivendell failed.")
        _logger.critical(e)
        return "Sending command to Rivendell failed."


def _send_email(arguments, log_name, desired_log_line, send_command_result):
    if arguments.emailer:
        email_subject = "[WMUL RIVENDELL] Workstation Loaded a Log from Script"
        email_body = f"Rivendell Workstation {arguments.rivendell_host} loaded a log from script.\n" \
                     f"The script was passed in the following arguments.\n" \
                     f"Datetime: {arguments.log_datetime.strftime('%Y-%m-%d  %H:%M:%S')}\n" \
                     f"Days Back: {arguments.days_back}\n" \
                     f"Default_Log: {arguments.default_log}\n" \
                     f"Log Machine: {arguments.log_machine}\n" \
                     f"Start Immediately: {arguments.start_immediately}\n\n" \
                     f"It tried to load {log_name}, and jump to line {desired_log_line[0]} with cart " \
                     f"{desired_log_line[1]} and start time {timedelta(milliseconds=int(desired_log_line[2]))}.\n" \
                     f"The result was: {send_command_result}"

        arguments.emailer.send_email(
            email_body=email_body,
            email_subject=email_subject,
        )


def run_script(arguments):
    _logger.debug(f"In run_script with {arguments}")
    desired_start_time_in_millis = _get_start_time_in_millis(arguments.log_datetime)
    _logger.info(f"Found desired_start_time_in_millis: {desired_start_time_in_millis}")
    desired_log_line, log_name = _get_desired_log_line(arguments, desired_start_time_in_millis)
    _logger.info(f"Found desired_log_line {desired_log_line} and log_name {log_name}")
    if desired_log_line:
        rml_command = _build_rml_command(desired_log_line, log_name, arguments.start_immediately, arguments.log_machine)
        _logger.info(f"Found rml_command: {rml_command}")
        if arguments.dry_run:
            print(f"Would have loaded log: {log_name} on host: {arguments.rivendell_host}, jumped to line "
                  f"{desired_log_line}, started_immediately: {arguments.start_immediately}, "
                  f"using the command: {rml_command}")
            return
        send_command_result = _send_command_to_rivendell(rml_command, arguments.rivendell_host)
    else:
        if arguments.dry_run:
            print(f"Unable to find a log line to load.")
            return
        send_command_result = "Could not run command, exhausted all possibilities"
    _send_email(arguments, log_name, desired_log_line, send_command_result)




"""
@Author = 'Michael Stanley'

This script takes the "Cart Data Dump (CSV)" from RD Library and converts it 
into Python data.


============ Change Log ============
2025-Jan-22 = Change debugging level on length_in_seconds.

2025-Jan-08 = Add logging to RivendellCart.from_dict, handle fields that may be string or NoneType.

              Add ability to fix csv lines that have un-escaped newlines.

              Add ability for RivendellCart.length_in_seconds to handle carts that are longer than 59:59.

2025-Jan-03 = Add repr for CartType enum.

              Strip extra chars when loading cart data dump, update tests.

              Remove the functionality for fixing the header bug from 3.6.4-3.6.6. It was a niche problem that is no 
              longer relevant and fixing it creates other problems.

              Add method to Rivendell Cart for converting the string MM:SS length to an int length_in_seconds.

2023-May-25 = Created. This module was originally part of 
                FilterCartReportForMusicSchedule and was refactored into this
                module so that DatabaseStatistics could re-use the code.

============ License ============
Copyright (C) 2020-2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import csv
from dataclasses import dataclass
from enum import Enum
from io import StringIO
from os import SEEK_SET
from pathlib import Path

import wmul_logger

_logger = wmul_logger.get_logger()



class CartType(Enum):
    Audio = 1
    Macro = 2

    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"


@dataclass
class RivendellCart:
    cart_number: str
    cut_number: str
    type: CartType
    group_name: str
    title: str
    artist: str
    album: str
    year: str
    isrc: str
    isci: str
    label: str
    client: str
    agency: str
    publisher: str
    composer: str
    conductor: str
    song_id: str
    user_defined: str
    description: str
    outcue: str
    filename: str
    length: str
    start_point: str
    end_point: str
    segue_start_point: str
    segue_end_point: str
    hook_start_point: str
    hook_end_point: str
    talk_start_point: str
    talk_end_point: str
    fadeup_point: str
    fadedown_point: str
    sched_codes: str

    @classmethod
    def from_dict(cls, source_dict):
        _logger.debug(source_dict)
        if source_dict["TYPE"] == "audio":
            cart_type = CartType.Audio
        else:
            cart_type = CartType.Macro

        description=source_dict["DESCRIPTION"]
        if isinstance(description, str):
            description = description.strip()

        outcue=source_dict["OUTCUE"]
        if isinstance(outcue, str):
            outcue = outcue.strip()

        return cls(
            cart_number=source_dict["CART_NUMBER"],
            cut_number=source_dict["CUT_NUMBER"],
            type=cart_type,
            group_name=source_dict["GROUP_NAME"],
            title=source_dict["TITLE"].strip(),
            artist=source_dict["ARTIST"].strip(),
            album=source_dict["ALBUM"].strip(),
            year=source_dict["YEAR"].strip(),
            isrc=source_dict["ISRC"].strip(),
            isci=source_dict["ISCI"].strip(),
            label=source_dict["LABEL"].strip(),
            client=source_dict["CLIENT"].strip(),
            agency=source_dict["AGENCY"].strip(),
            publisher=source_dict["PUBLISHER"].strip(),
            composer=source_dict["COMPOSER"].strip(),
            conductor=source_dict["CONDUCTOR"].strip(),
            song_id=source_dict["SONG_ID"].strip(),
            user_defined=source_dict["USER_DEFINED"].strip(),
            description=description,
            outcue=outcue,
            filename=source_dict["FILENAME"],
            length=source_dict["LENGTH"],
            start_point=source_dict["START_POINT"],
            end_point=source_dict["END_POINT"],
            segue_start_point=source_dict["SEGUE_START_POINT"],
            segue_end_point=source_dict["SEGUE_END_POINT"],
            hook_start_point=source_dict["HOOK_START_POINT"],
            hook_end_point=source_dict["HOOK_END_POINT"],
            talk_start_point=source_dict["TALK_START_POINT"],
            talk_end_point=source_dict["TALK_END_POINT"],
            fadeup_point=source_dict["FADEUP_POINT"],
            fadedown_point=source_dict["FADEDOWN_POINT"],
            sched_codes=source_dict["SCHED_CODES"]
        )

    def length_in_seconds(self):
        _logger.debug(f"{self.group_name}:{self.cart_number} {self.length}")
        length = self.length
        if length.count(":") > 1:
            hours, minutes, seconds = length.split(":")
            hours = int(hours)
        else:
            minutes, seconds = self.length.split(":")
            hours = 0

        if minutes:
            minutes = int(minutes)
        else:
            minutes = 0
        seconds = int(seconds)

        return (hours * 3600) + (minutes * 60) + seconds


def _fix_rivendell_csv_file(rivendell_source_file):
    text_buffer = StringIO()

    file_iterator = iter(rivendell_source_file)
    expected_fields = next(file_iterator)
    text_buffer.write(expected_fields)

    line_buffer = StringIO(expected_fields)
    line_parsed_as_csv = list(csv.reader(line_buffer))[0]

    count_of_expected_fields = len(line_parsed_as_csv)

    try:
        while True:
            text_line = next(file_iterator)
            text_line_buffer = StringIO(text_line)
            text_line_parsed_as_csv = list(csv.reader(text_line_buffer))[0]
            
            while len(text_line_parsed_as_csv) < count_of_expected_fields:
                text_line = text_line.strip()
                next_line = next(file_iterator)
                text_line = text_line + " " + next_line
                text_line_buffer = StringIO(text_line)
                text_line_parsed_as_csv = list(csv.reader(text_line_buffer))[0]

                if len(text_line_parsed_as_csv) > count_of_expected_fields:
                    raise ValueError(f"There are errors in the csv file that cannot be automatically fixed. {text_line_buffer.getvalue()}")
            text_buffer.write(text_line_buffer.getvalue())
    except StopIteration:
        text_buffer.seek(0, SEEK_SET)
        return text_buffer


@dataclass
class LoadCartDataDump:
    rivendell_cart_data_filename: Path
    excluded_group_list: list
    include_macros: bool
    include_all_cuts: bool

    def _load_rivendell_carts(self):
        with open(str(self.rivendell_cart_data_filename), newline="", mode="rt", errors="replace") as \
                rivendell_source_file:
            rivendell_fixed_data = _fix_rivendell_csv_file(rivendell_source_file)
            rivendell_reader = csv.DictReader(rivendell_fixed_data)
            rivendell_carts = [RivendellCart.from_dict(rivendell_cart) for rivendell_cart in rivendell_reader]
        return rivendell_carts

    def _remove_excluded_groups(self, rivendell_carts):
        return (rivendell_cart for rivendell_cart in rivendell_carts if
                rivendell_cart.group_name not in self.excluded_group_list)

    def _remove_macro_carts(self, rivendell_carts):
        return (rivendell_cart for rivendell_cart in rivendell_carts if not rivendell_cart.type == CartType.Macro)

    def _remove_extra_cuts(self, rivendell_carts):
        carts_grouped_by_cart_number = dict()
        for cart in rivendell_carts:
            if cart.cart_number in carts_grouped_by_cart_number:
                if cart.cut_number < carts_grouped_by_cart_number[cart.cart_number].cut_number:
                    carts_grouped_by_cart_number[cart.cart_number] = cart
            else:
                carts_grouped_by_cart_number[cart.cart_number] = cart
        return carts_grouped_by_cart_number.values()

    def load_carts(self):
        _logger.debug(f"Starting load_carts with {self}")
        rivendell_carts = self._load_rivendell_carts()

        if self.excluded_group_list:
            rivendell_carts = self._remove_excluded_groups(rivendell_carts)
        if not self.include_macros:
            rivendell_carts = self._remove_macro_carts(rivendell_carts)
        if not self.include_all_cuts:
            rivendell_carts = self._remove_extra_cuts(rivendell_carts)

        return rivendell_carts



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2022-May-16 = Fix a bug caused by backporting behind Python 3.7, which does 
                not have the subprocess.run capture_output param.

2020-Oct-08 = Created.

============ License ============
Copyright (C) 2020, 2022-2023 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import pytest
import wmul_test_utils

from wmul_rivendell import RivendellAudioImporter


@pytest.fixture(scope="function")
def setup__run_importer_on_selected_files(caplog, mocker):
    mock_file_1_command = "mock_file_1_command"
    mock_file_2_command = "mock_file_2_command"
    mock_file_3_command = "mock_file_3_command"
    mock_file_4_command = "mock_file_4_command"

    mock_file_1_file_path = "mock_file_1_file_path"
    mock_file_2_file_path = "mock_file_2_file_path"
    mock_file_3_file_path = "mock_file_3_file_path"
    mock_file_4_file_path = "mock_file_4_file_path"

    mock_file_1_failed = mocker.Mock()
    mock_file_2_failed = mocker.Mock()
    mock_file_3_failed = mocker.Mock()
    mock_file_4_failed = mocker.Mock()

    mock_file_1 = mocker.Mock(
        generate_importer_command=mocker.Mock(return_value=[mock_file_1_command]),
        file_path=mock_file_1_file_path,
        rdimport_failed=mock_file_1_failed
    )
    mock_file_2 = mocker.Mock(
        generate_importer_command=mocker.Mock(return_value=[mock_file_2_command]),
        file_path=mock_file_2_file_path,
        rdimport_failed=mock_file_2_failed
    )
    mock_file_3 = mocker.Mock(
        generate_importer_command=mocker.Mock(return_value=[mock_file_3_command]),
        file_path=mock_file_3_file_path,
        rdimport_failed=mock_file_3_failed
    )
    mock_file_4 = mocker.Mock(
        generate_importer_command=mocker.Mock(return_value=[mock_file_4_command]),
        file_path=mock_file_4_file_path,
        rdimport_failed=mock_file_4_failed
    )

    mock_files_for_importer = [mock_file_1, mock_file_2, mock_file_3, mock_file_4]

    mock_previously_sent_to_importer = {}

    mock_log_arguement = "mock_log_argument"

    file_4_stdout = b'blah blah is not readable or not a recognized format, skipping... blah blah'

    def mock_subprocess_run_function(args, stdout, stderr):
        if args[1] == mock_file_2_command:
            return mocker.Mock(returncode=-1, args=args, stdout=b'', stderr=b'')
        elif args[1] == mock_file_4_command:
            return mocker.Mock(
                returncode=0,
                args=args,
                stdout=file_4_stdout,
                stderr=b''
            )
        else:
            return mocker.Mock(returncode=0, args=args, stdout=b'', stderr=b'')

    mock_subprocess_run = mocker.Mock(side_effect=mock_subprocess_run_function)
    mocker.patch("wmul_rivendell.RivendellAudioImporter.subprocess.run",
                 mock_subprocess_run)

    mock_subprocess_pipe = "mock_suprocess_pipe"
    mocker.patch("wmul_rivendell.RivendellAudioImporter.subprocess.PIPE",
                 mock_subprocess_pipe)

    RivendellAudioImporter._run_importer_on_selected_files(
        mock_files_for_importer,
        mock_previously_sent_to_importer,
        mock_log_arguement
    )

    expected_calls = [
        mocker.call(["rdimport", mock_file_1_command], stderr=mock_subprocess_pipe, stdout=mock_subprocess_pipe),
        mocker.call(["rdimport", mock_file_2_command], stderr=mock_subprocess_pipe, stdout=mock_subprocess_pipe),
        mocker.call(["rdimport", mock_file_3_command], stderr=mock_subprocess_pipe, stdout=mock_subprocess_pipe),
        mocker.call(["rdimport", mock_file_4_command], stderr=mock_subprocess_pipe, stdout=mock_subprocess_pipe)
    ]

    expected_previously_sent_to_importer = {
        mock_file_1_file_path: True,
        mock_file_2_file_path: True,
        mock_file_3_file_path: True,
        mock_file_4_file_path: True
    }

    expected_log_entries = [
        f"Sending to the importer: {mock_file_1_file_path}",
        f"Importer command: ['{mock_file_1_command}']",
        f"Success! {mock_file_1_command}, ['rdimport', '{mock_file_1_command}'], b'', b''",
        f"Sending to the importer: {mock_file_2_file_path}",
        f"Importer command: ['{mock_file_2_command}']",
        f"Error on {mock_file_2_command}, ['rdimport', '{mock_file_2_command}'], b'', b''",
        f"Sending to the importer: {mock_file_3_file_path}",
        f"Importer command: ['{mock_file_3_command}']",
        f"Success! {mock_file_3_command}, ['rdimport', '{mock_file_3_command}'], b'', b''",
        f"Sending to the importer: {mock_file_4_file_path}",
        f"Importer command: ['{mock_file_4_command}']",
        f"Error on {mock_file_4_command}, ['rdimport', '{mock_file_4_command}'], b'', {file_4_stdout}",
    ]

    return wmul_test_utils.make_namedtuple(
        "setup__run_importer_on_selected_files",
        mock_files_for_importer=mock_files_for_importer,
        mock_previously_sent_to_importer=mock_previously_sent_to_importer,
        mock_log_arguement=mock_log_arguement,
        mock_subprocess_run=mock_subprocess_run,
        expected_calls=expected_calls,
        expected_previously_sent_to_importer=expected_previously_sent_to_importer,
        expected_log_entries=expected_log_entries,
        result_log_entries=caplog.messages
    )


def test_generate_importer_command_called_correctly(
        setup__run_importer_on_selected_files):
    for mock_file in setup__run_importer_on_selected_files.mock_files_for_importer:
        mock_file.generate_importer_command.assert_called_once_with(
            setup__run_importer_on_selected_files.mock_log_arguement
        )


def test_cache_correct(setup__run_importer_on_selected_files):
    assert setup__run_importer_on_selected_files.mock_previously_sent_to_importer == \
           setup__run_importer_on_selected_files.expected_previously_sent_to_importer


def test_subprocess_called_correctly(setup__run_importer_on_selected_files):
    setup__run_importer_on_selected_files.mock_subprocess_run.assert_has_calls(
        setup__run_importer_on_selected_files.expected_calls)

    assert setup__run_importer_on_selected_files.mock_subprocess_run.call_count == \
           len(setup__run_importer_on_selected_files.expected_calls)


def test_logging_correct(setup__run_importer_on_selected_files):
    assert sorted(setup__run_importer_on_selected_files.expected_log_entries) == \
           sorted(setup__run_importer_on_selected_files.result_log_entries)



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2022-May-05 = Make .wav case insensitive.

2020-Oct-07 = Created.

============ License ============
Copyright (C) 2020, 2022-2023 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import cachetools
import pytest
import wmul_test_utils

from wmul_rivendell import RivendellAudioImporter


@pytest.fixture(scope="function")
def setup__gather_names_from_this_path(mocker, caplog):
    mock_previously_seen_files = "mock_previously_seen_files"
    mock_datetime_now = "mock_datetime_now"

    mock_source_path_1 = "mock_source_path_1"
    mock_source_path_2 = "mock_source_path_2"
    mock_source_path_3 = "mock_source_path_3"
    mock_source_path_4 = "mock_source_path_4"

    mock_ttl = 40_000
    mock_previously_sent_to_importer_cache = cachetools.TTLCache(
        10_000,
        mock_ttl
    )
    mock_previously_sent_to_importer_cache[mock_source_path_4] = True

    mock_source_path_rglob = [
        mock_source_path_1,
        mock_source_path_2,
        mock_source_path_3,
        mock_source_path_4
    ]

    mock_source_path_rglob_function = mocker.Mock(return_value=mock_source_path_rglob)

    mock_source_path = mocker.Mock(rglob=mock_source_path_rglob_function)

    mock_seen_this_time = mocker.Mock(return_value=True)

    def mock__check_this_file_function(psf, fp, dtn, sp):
        if fp == mock_source_path_1:
            return None, mock_source_path_1
        elif fp == mock_source_path_2:
            return None, None
        elif fp == mock_source_path_3:
            return mock_seen_this_time, None
        else:
            raise ValueError("file path not valid")

    mock__check_this_file = mocker.Mock(side_effect=mock__check_this_file_function)

    mocker.patch("wmul_rivendell.RivendellAudioImporter._check_this_file", mock__check_this_file)

    expected_files_seen_this_time = {mock_source_path_3: mock_seen_this_time}
    expected_files_for_importer = [mock_source_path_1]

    result_files_seen_this_time, result_files_for_importer = \
        RivendellAudioImporter._gather_names_from_this_path(
            mock_previously_seen_files,
            mock_datetime_now,
            mock_source_path,
            mock_previously_sent_to_importer_cache
        )

    expected_calls = [
        mocker.call(mock_previously_seen_files, mock_source_path_rglob[0], mock_datetime_now, mock_source_path),
        mocker.call(mock_previously_seen_files, mock_source_path_rglob[1], mock_datetime_now, mock_source_path),
        mocker.call(mock_previously_seen_files, mock_source_path_rglob[2], mock_datetime_now, mock_source_path)
    ]

    expected_log_messages = [
        f"With Previously Seen Files: {mock_previously_seen_files}, Datetime_now: {mock_datetime_now}, "
        f"Source Path: {mock_source_path}, and Previously Sent to Importer Cache: "
        f"{mock_previously_sent_to_importer_cache}",

        f"Working on {mock_source_path_1}",
        f"Working on {mock_source_path_2}",
        f"Working on {mock_source_path_3}",
        f"Working on {mock_source_path_4}",
        f"Have not sent {mock_source_path_1} to the importer in the previous {mock_ttl} seconds",
        f"Have not sent {mock_source_path_2} to the importer in the previous {mock_ttl} seconds",
        f"Have not sent {mock_source_path_3} to the importer in the previous {mock_ttl} seconds",
        f"Seen this time! {mock_source_path_3}",
        f"For importer! {mock_source_path_1}",
        f"Not for seen this time or importer! {mock_source_path_2}",
        f"Sent a file with this name to the importer within the previous {mock_ttl} seconds! {mock_source_path_4}",
        f"Returning files_seen_this_time: {expected_files_seen_this_time} and "
        f"files_for_importer {expected_files_for_importer}"
    ]

    return wmul_test_utils.make_namedtuple(
        "setup__gather_names_from_this_path",
        mock_source_path_rglob_function=mock_source_path_rglob_function,
        mock__check_this_file=mock__check_this_file,
        expected_files_seen_this_time=expected_files_seen_this_time,
        expected_files_for_importer=expected_files_for_importer,
        result_files_seen_this_time=result_files_seen_this_time,
        result_files_for_importer=result_files_for_importer,
        expected_calls=expected_calls,
        expected_log_messages=expected_log_messages,
        result_log_messages=caplog.messages
    )


def test_source_path_rglob_called_correctly(setup__gather_names_from_this_path):
    setup__gather_names_from_this_path.mock_source_path_rglob_function.assert_called_once_with("*.[Ww][Aa][Vv]")


def test__check_this_file_called_correctly(setup__gather_names_from_this_path):
    setup__gather_names_from_this_path.mock__check_this_file.assert_has_calls(
        setup__gather_names_from_this_path.expected_calls)

    assert setup__gather_names_from_this_path.mock__check_this_file.call_count == \
           len(setup__gather_names_from_this_path.expected_calls)


def test_files_seen_this_time_correct(setup__gather_names_from_this_path):
    assert setup__gather_names_from_this_path.expected_files_seen_this_time == \
           setup__gather_names_from_this_path.result_files_seen_this_time


def test_files_for_importer_correct(setup__gather_names_from_this_path):
    assert setup__gather_names_from_this_path.expected_files_for_importer == \
           setup__gather_names_from_this_path.result_files_for_importer


def test_log_messages_correct(setup__gather_names_from_this_path):
    assert sorted(setup__gather_names_from_this_path.expected_log_messages) == \
           sorted(setup__gather_names_from_this_path.result_log_messages)
